/* DO NOT EDIT THIS FILE - it is machine generated */
#include <jni.h>
/* Header for class jportmidi_JPortMidiApi */

#ifndef _Included_jportmidi_JPortMidiApi
#define _Included_jportmidi_JPortMidiApi
#ifdef __cplusplus
extern "C" {
#endif
#undef jportmidi_JPortMidiApi_PM_FILT_ACTIVE
#define jportmidi_JPortMidiApi_PM_FILT_ACTIVE 16384L
#undef jportmidi_JPortMidiApi_PM_FILT_SYSEX
#define jportmidi_JPortMidiApi_PM_FILT_SYSEX 1L
#undef jportmidi_JPortMidiApi_PM_FILT_CLOCK
#define jportmidi_JPortMidiApi_PM_FILT_CLOCK 256L
#undef jportmidi_JPortMidiApi_PM_FILT_PLAY
#define jportmidi_JPortMidiApi_PM_FILT_PLAY 7168L
#undef jportmidi_JPortMidiApi_PM_FILT_TICK
#define jportmidi_JPortMidiApi_PM_FILT_TICK 512L
#undef jportmidi_JPortMidiApi_PM_FILT_FD
#define jportmidi_JPortMidiApi_PM_FILT_FD 8192L
#undef jportmidi_JPortMidiApi_PM_FILT_UNDEFINED
#define jportmidi_JPortMidiApi_PM_FILT_UNDEFINED 8192L
#undef jportmidi_JPortMidiApi_PM_FILT_RESET
#define jportmidi_JPortMidiApi_PM_FILT_RESET 32768L
#undef jportmidi_JPortMidiApi_PM_FILT_REALTIME
#define jportmidi_JPortMidiApi_PM_FILT_REALTIME 16641L
#undef jportmidi_JPortMidiApi_PM_FILT_NOTE
#define jportmidi_JPortMidiApi_PM_FILT_NOTE 50331648L
#undef jportmidi_JPortMidiApi_PM_FILT_CHANNEL_AFTERTOUCH
#define jportmidi_JPortMidiApi_PM_FILT_CHANNEL_AFTERTOUCH 536870912L
#undef jportmidi_JPortMidiApi_PM_FILT_POLY_AFTERTOUCH
#define jportmidi_JPortMidiApi_PM_FILT_POLY_AFTERTOUCH 67108864L
#undef jportmidi_JPortMidiApi_PM_FILT_AFTERTOUCH
#define jportmidi_JPortMidiApi_PM_FILT_AFTERTOUCH 603979776L
#undef jportmidi_JPortMidiApi_PM_FILT_PROGRAM
#define jportmidi_JPortMidiApi_PM_FILT_PROGRAM 268435456L
#undef jportmidi_JPortMidiApi_PM_FILT_CONTROL
#define jportmidi_JPortMidiApi_PM_FILT_CONTROL 134217728L
#undef jportmidi_JPortMidiApi_PM_FILT_PITCHBEND
#define jportmidi_JPortMidiApi_PM_FILT_PITCHBEND 1073741824L
#undef jportmidi_JPortMidiApi_PM_FILT_MTC
#define jportmidi_JPortMidiApi_PM_FILT_MTC 2L
#undef jportmidi_JPortMidiApi_PM_FILT_SONG_POSITION
#define jportmidi_JPortMidiApi_PM_FILT_SONG_POSITION 4L
#undef jportmidi_JPortMidiApi_PM_FILT_SONG_SELECT
#define jportmidi_JPortMidiApi_PM_FILT_SONG_SELECT 8L
#undef jportmidi_JPortMidiApi_PM_FILT_TUNE
#define jportmidi_JPortMidiApi_PM_FILT_TUNE 64L
#undef jportmidi_JPortMidiApi_PM_FILT_SYSTEMCOMMON
#define jportmidi_JPortMidiApi_PM_FILT_SYSTEMCOMMON 78L
/*
 * Class:     jportmidi_JPortMidiApi
 * Method:    Pm_Initialize
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_jportmidi_JPortMidiApi_Pm_1Initialize
  (JNIEnv *, jclass);

/*
 * Class:     jportmidi_JPortMidiApi
 * Method:    Pm_Terminate
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_jportmidi_JPortMidiApi_Pm_1Terminate
  (JNIEnv *, jclass);

/*
 * Class:     jportmidi_JPortMidiApi
 * Method:    Pm_HasHostError
 * Signature: (Ljportmidi/JPortMidiApi/PortMidiStream;)I
 */
JNIEXPORT jint JNICALL Java_jportmidi_JPortMidiApi_Pm_1HasHostError
  (JNIEnv *, jclass, jobject);

/*
 * Class:     jportmidi_JPortMidiApi
 * Method:    Pm_GetErrorText
 * Signature: (I)Ljava/lang/String;
 */
JNIEXPORT jstring JNICALL Java_jportmidi_JPortMidiApi_Pm_1GetErrorText
  (JNIEnv *, jclass, jint);

/*
 * Class:     jportmidi_JPortMidiApi
 * Method:    Pm_GetHostErrorText
 * Signature: ()Ljava/lang/String;
 */
JNIEXPORT jstring JNICALL Java_jportmidi_JPortMidiApi_Pm_1GetHostErrorText
  (JNIEnv *, jclass);

/*
 * Class:     jportmidi_JPortMidiApi
 * Method:    Pm_CountDevices
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_jportmidi_JPortMidiApi_Pm_1CountDevices
  (JNIEnv *, jclass);

/*
 * Class:     jportmidi_JPortMidiApi
 * Method:    Pm_GetDefaultInputDeviceID
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_jportmidi_JPortMidiApi_Pm_1GetDefaultInputDeviceID
  (JNIEnv *, jclass);

/*
 * Class:     jportmidi_JPortMidiApi
 * Method:    Pm_GetDefaultOutputDeviceID
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_jportmidi_JPortMidiApi_Pm_1GetDefaultOutputDeviceID
  (JNIEnv *, jclass);

/*
 * Class:     jportmidi_JPortMidiApi
 * Method:    Pm_GetDeviceInterf
 * Signature: (I)Ljava/lang/String;
 */
JNIEXPORT jstring JNICALL Java_jportmidi_JPortMidiApi_Pm_1GetDeviceInterf
  (JNIEnv *, jclass, jint);

/*
 * Class:     jportmidi_JPortMidiApi
 * Method:    Pm_GetDeviceName
 * Signature: (I)Ljava/lang/String;
 */
JNIEXPORT jstring JNICALL Java_jportmidi_JPortMidiApi_Pm_1GetDeviceName
  (JNIEnv *, jclass, jint);

/*
 * Class:     jportmidi_JPortMidiApi
 * Method:    Pm_GetDeviceInput
 * Signature: (I)Z
 */
JNIEXPORT jboolean JNICALL Java_jportmidi_JPortMidiApi_Pm_1GetDeviceInput
  (JNIEnv *, jclass, jint);

/*
 * Class:     jportmidi_JPortMidiApi
 * Method:    Pm_GetDeviceOutput
 * Signature: (I)Z
 */
JNIEXPORT jboolean JNICALL Java_jportmidi_JPortMidiApi_Pm_1GetDeviceOutput
  (JNIEnv *, jclass, jint);

/*
 * Class:     jportmidi_JPortMidiApi
 * Method:    Pm_OpenInput
 * Signature: (Ljportmidi/JPortMidiApi/PortMidiStream;ILjava/lang/String;I)I
 */
JNIEXPORT jint JNICALL Java_jportmidi_JPortMidiApi_Pm_1OpenInput
  (JNIEnv *, jclass, jobject, jint, jstring, jint);

/*
 * Class:     jportmidi_JPortMidiApi
 * Method:    Pm_OpenOutput
 * Signature: (Ljportmidi/JPortMidiApi/PortMidiStream;ILjava/lang/String;II)I
 */
JNIEXPORT jint JNICALL Java_jportmidi_JPortMidiApi_Pm_1OpenOutput
  (JNIEnv *, jclass, jobject, jint, jstring, jint, jint);

/*
 * Class:     jportmidi_JPortMidiApi
 * Method:    Pm_SetFilter
 * Signature: (Ljportmidi/JPortMidiApi/PortMidiStream;I)I
 */
JNIEXPORT jint JNICALL Java_jportmidi_JPortMidiApi_Pm_1SetFilter
  (JNIEnv *, jclass, jobject, jint);

/*
 * Class:     jportmidi_JPortMidiApi
 * Method:    Pm_SetChannelMask
 * Signature: (Ljportmidi/JPortMidiApi/PortMidiStream;I)I
 */
JNIEXPORT jint JNICALL Java_jportmidi_JPortMidiApi_Pm_1SetChannelMask
  (JNIEnv *, jclass, jobject, jint);

/*
 * Class:     jportmidi_JPortMidiApi
 * Method:    Pm_Abort
 * Signature: (Ljportmidi/JPortMidiApi/PortMidiStream;)I
 */
JNIEXPORT jint JNICALL Java_jportmidi_JPortMidiApi_Pm_1Abort
  (JNIEnv *, jclass, jobject);

/*
 * Class:     jportmidi_JPortMidiApi
 * Method:    Pm_Close
 * Signature: (Ljportmidi/JPortMidiApi/PortMidiStream;)I
 */
JNIEXPORT jint JNICALL Java_jportmidi_JPortMidiApi_Pm_1Close
  (JNIEnv *, jclass, jobject);

/*
 * Class:     jportmidi_JPortMidiApi
 * Method:    Pm_Read
 * Signature: (Ljportmidi/JPortMidiApi/PortMidiStream;Ljportmidi/JPortMidiApi/PmEvent;)I
 */
JNIEXPORT jint JNICALL Java_jportmidi_JPortMidiApi_Pm_1Read
  (JNIEnv *, jclass, jobject, jobject);

/*
 * Class:     jportmidi_JPortMidiApi
 * Method:    Pm_Poll
 * Signature: (Ljportmidi/JPortMidiApi/PortMidiStream;)I
 */
JNIEXPORT jint JNICALL Java_jportmidi_JPortMidiApi_Pm_1Poll
  (JNIEnv *, jclass, jobject);

/*
 * Class:     jportmidi_JPortMidiApi
 * Method:    Pm_Write
 * Signature: (Ljportmidi/JPortMidiApi/PortMidiStream;Ljportmidi/JPortMidiApi/PmEvent;)I
 */
JNIEXPORT jint JNICALL Java_jportmidi_JPortMidiApi_Pm_1Write
  (JNIEnv *, jclass, jobject, jobject);

/*
 * Class:     jportmidi_JPortMidiApi
 * Method:    Pm_WriteShort
 * Signature: (Ljportmidi/JPortMidiApi/PortMidiStream;II)I
 */
JNIEXPORT jint JNICALL Java_jportmidi_JPortMidiApi_Pm_1WriteShort
  (JNIEnv *, jclass, jobject, jint, jint);

/*
 * Class:     jportmidi_JPortMidiApi
 * Method:    Pm_WriteSysEx
 * Signature: (Ljportmidi/JPortMidiApi/PortMidiStream;I[B)I
 */
JNIEXPORT jint JNICALL Java_jportmidi_JPortMidiApi_Pm_1WriteSysEx
  (JNIEnv *, jclass, jobject, jint, jbyteArray);

/*
 * Class:     jportmidi_JPortMidiApi
 * Method:    Pt_TimeStart
 * Signature: (I)I
 */
JNIEXPORT jint JNICALL Java_jportmidi_JPortMidiApi_Pt_1TimeStart
  (JNIEnv *, jclass, jint);

/*
 * Class:     jportmidi_JPortMidiApi
 * Method:    Pt_TimeStop
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_jportmidi_JPortMidiApi_Pt_1TimeStop
  (JNIEnv *, jclass);

/*
 * Class:     jportmidi_JPortMidiApi
 * Method:    Pt_Time
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_jportmidi_JPortMidiApi_Pt_1Time
  (JNIEnv *, jclass);

/*
 * Class:     jportmidi_JPortMidiApi
 * Method:    Pt_TimeStarted
 * Signature: ()Z
 */
JNIEXPORT jboolean JNICALL Java_jportmidi_JPortMidiApi_Pt_1TimeStarted
  (JNIEnv *, jclass);

#ifdef __cplusplus
}
#endif
#endif
/* Header for class jportmidi_JPortMidiApi_PmEvent */

#ifndef _Included_jportmidi_JPortMidiApi_PmEvent
#define _Included_jportmidi_JPortMidiApi_PmEvent
#ifdef __cplusplus
extern "C" {
#endif
#ifdef __cplusplus
}
#endif
#endif
/* Header for class jportmidi_JPortMidiApi_PortMidiStream */

#ifndef _Included_jportmidi_JPortMidiApi_PortMidiStream
#define _Included_jportmidi_JPortMidiApi_PortMidiStream
#ifdef __cplusplus
extern "C" {
#endif
#ifdef __cplusplus
}
#endif
#endif
