// Emacs style mode select   -*- C++ -*-
//-----------------------------------------------------------------------------
//
// $Id: 62b332d8206004f8eb375544a1f4533c5c7a3636 $
//
// Copyright (C) 1998-2011 by Randy Heit (ZDoom 1.23).
// Copyright (C) 2006-2026 by The Odamex Team.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// DESCRIPTION:
// Defines names for the indices in the string table.
// This file is automatically generated.
//
//-----------------------------------------------------------------------------

#pragma once

#include "m_ostring.h"

inline const OString D_DEVSTR = "D_DEVSTR";
inline const OString D_CDROM = "D_CDROM";
inline const OString PRESSKEY = "PRESSKEY";
inline const OString PRESSYN = "PRESSYN";
inline const OString QUITMSG = "QUITMSG";
inline const OString QUITMSG1 = "QUITMSG1";
inline const OString QUITMSG2 = "QUITMSG2";
inline const OString QUITMSG3 = "QUITMSG3";
inline const OString QUITMSG4 = "QUITMSG4";
inline const OString QUITMSG5 = "QUITMSG5";
inline const OString QUITMSG6 = "QUITMSG6";
inline const OString QUITMSG7 = "QUITMSG7";
inline const OString QUITMSG8 = "QUITMSG8";
inline const OString QUITMSG9 = "QUITMSG9";
inline const OString QUITMSG10 = "QUITMSG10";
inline const OString QUITMSG11 = "QUITMSG11";
inline const OString QUITMSG12 = "QUITMSG12";
inline const OString QUITMSG13 = "QUITMSG13";
inline const OString QUITMSG14 = "QUITMSG14";
inline const OString LOADNET = "LOADNET";
inline const OString QLOADNET = "QLOADNET";
inline const OString QSAVESPOT = "QSAVESPOT";
inline const OString SAVEDEAD = "SAVEDEAD";
inline const OString QSPROMPT = "QSPROMPT";
inline const OString QLPROMPT = "QLPROMPT";
inline const OString NEWGAME = "NEWGAME";
inline const OString NIGHTMARE = "NIGHTMARE";
inline const OString SWSTRING = "SWSTRING";
inline const OString MSGOFF = "MSGOFF";
inline const OString MSGON = "MSGON";
inline const OString NETEND = "NETEND";
inline const OString ENDGAME = "ENDGAME";
inline const OString DOSY = "DOSY";
inline const OString EMPTYSTRING = "EMPTYSTRING";
inline const OString GOTARMOR = "GOTARMOR";
inline const OString GOTMEGA = "GOTMEGA";
inline const OString GOTHTHBONUS = "GOTHTHBONUS";
inline const OString GOTARMBONUS = "GOTARMBONUS";
inline const OString GOTSTIM = "GOTSTIM";
inline const OString GOTMEDINEED = "GOTMEDINEED";
inline const OString GOTMEDIKIT = "GOTMEDIKIT";
inline const OString GOTSUPER = "GOTSUPER";
inline const OString GOTBLUECARD = "GOTBLUECARD";
inline const OString GOTYELWCARD = "GOTYELWCARD";
inline const OString GOTREDCARD = "GOTREDCARD";
inline const OString GOTBLUESKUL = "GOTBLUESKUL";
inline const OString GOTYELWSKUL = "GOTYELWSKUL";
inline const OString GOTREDSKUL = "GOTREDSKUL";
inline const OString GOTINVUL = "GOTINVUL";
inline const OString GOTBERSERK = "GOTBERSERK";
inline const OString GOTINVIS = "GOTINVIS";
inline const OString GOTSUIT = "GOTSUIT";
inline const OString GOTMAP = "GOTMAP";
inline const OString GOTVISOR = "GOTVISOR";
inline const OString GOTMSPHERE = "GOTMSPHERE";
inline const OString GOTCLIP = "GOTCLIP";
inline const OString GOTCLIPBOX = "GOTCLIPBOX";
inline const OString GOTROCKET = "GOTROCKET";
inline const OString GOTROCKBOX = "GOTROCKBOX";
inline const OString GOTCELL = "GOTCELL";
inline const OString GOTCELLBOX = "GOTCELLBOX";
inline const OString GOTSHELLS = "GOTSHELLS";
inline const OString GOTSHELLBOX = "GOTSHELLBOX";
inline const OString GOTBACKPACK = "GOTBACKPACK";
inline const OString GOTBFG9000 = "GOTBFG9000";
inline const OString GOTCHAINGUN = "GOTCHAINGUN";
inline const OString GOTCHAINSAW = "GOTCHAINSAW";
inline const OString GOTLAUNCHER = "GOTLAUNCHER";
inline const OString GOTPLASMA = "GOTPLASMA";
inline const OString GOTSHOTGUN = "GOTSHOTGUN";
inline const OString GOTSHOTGUN2 = "GOTSHOTGUN2";
inline const OString PD_BLUEO = "PD_BLUEO";
inline const OString PD_REDO = "PD_REDO";
inline const OString PD_YELLOWO = "PD_YELLOWO";
inline const OString PD_BLUEK = "PD_BLUEK";
inline const OString PD_REDK = "PD_REDK";
inline const OString PD_YELLOWK = "PD_YELLOWK";
inline const OString GGSAVED = "GGSAVED";
inline const OString HUSTR_MSGU = "HUSTR_MSGU";
inline const OString HUSTR_E1M1 = "HUSTR_E1M1";
inline const OString HUSTR_E1M2 = "HUSTR_E1M2";
inline const OString HUSTR_E1M3 = "HUSTR_E1M3";
inline const OString HUSTR_E1M4 = "HUSTR_E1M4";
inline const OString HUSTR_E1M5 = "HUSTR_E1M5";
inline const OString HUSTR_E1M6 = "HUSTR_E1M6";
inline const OString HUSTR_E1M7 = "HUSTR_E1M7";
inline const OString HUSTR_E1M8 = "HUSTR_E1M8";
inline const OString HUSTR_E1M9 = "HUSTR_E1M9";
inline const OString HUSTR_E2M1 = "HUSTR_E2M1";
inline const OString HUSTR_E2M2 = "HUSTR_E2M2";
inline const OString HUSTR_E2M3 = "HUSTR_E2M3";
inline const OString HUSTR_E2M4 = "HUSTR_E2M4";
inline const OString HUSTR_E2M5 = "HUSTR_E2M5";
inline const OString HUSTR_E2M6 = "HUSTR_E2M6";
inline const OString HUSTR_E2M7 = "HUSTR_E2M7";
inline const OString HUSTR_E2M8 = "HUSTR_E2M8";
inline const OString HUSTR_E2M9 = "HUSTR_E2M9";
inline const OString HUSTR_E3M1 = "HUSTR_E3M1";
inline const OString HUSTR_E3M2 = "HUSTR_E3M2";
inline const OString HUSTR_E3M3 = "HUSTR_E3M3";
inline const OString HUSTR_E3M4 = "HUSTR_E3M4";
inline const OString HUSTR_E3M5 = "HUSTR_E3M5";
inline const OString HUSTR_E3M6 = "HUSTR_E3M6";
inline const OString HUSTR_E3M7 = "HUSTR_E3M7";
inline const OString HUSTR_E3M8 = "HUSTR_E3M8";
inline const OString HUSTR_E3M9 = "HUSTR_E3M9";
inline const OString HUSTR_E4M1 = "HUSTR_E4M1";
inline const OString HUSTR_E4M2 = "HUSTR_E4M2";
inline const OString HUSTR_E4M3 = "HUSTR_E4M3";
inline const OString HUSTR_E4M4 = "HUSTR_E4M4";
inline const OString HUSTR_E4M5 = "HUSTR_E4M5";
inline const OString HUSTR_E4M6 = "HUSTR_E4M6";
inline const OString HUSTR_E4M7 = "HUSTR_E4M7";
inline const OString HUSTR_E4M8 = "HUSTR_E4M8";
inline const OString HUSTR_E4M9 = "HUSTR_E4M9";
inline const OString HUSTR_1 = "HUSTR_1";
inline const OString HUSTR_2 = "HUSTR_2";
inline const OString HUSTR_3 = "HUSTR_3";
inline const OString HUSTR_4 = "HUSTR_4";
inline const OString HUSTR_5 = "HUSTR_5";
inline const OString HUSTR_6 = "HUSTR_6";
inline const OString HUSTR_7 = "HUSTR_7";
inline const OString HUSTR_8 = "HUSTR_8";
inline const OString HUSTR_9 = "HUSTR_9";
inline const OString HUSTR_10 = "HUSTR_10";
inline const OString HUSTR_11 = "HUSTR_11";
inline const OString HUSTR_12 = "HUSTR_12";
inline const OString HUSTR_13 = "HUSTR_13";
inline const OString HUSTR_14 = "HUSTR_14";
inline const OString HUSTR_15 = "HUSTR_15";
inline const OString HUSTR_16 = "HUSTR_16";
inline const OString HUSTR_17 = "HUSTR_17";
inline const OString HUSTR_18 = "HUSTR_18";
inline const OString HUSTR_19 = "HUSTR_19";
inline const OString HUSTR_20 = "HUSTR_20";
inline const OString HUSTR_21 = "HUSTR_21";
inline const OString HUSTR_22 = "HUSTR_22";
inline const OString HUSTR_23 = "HUSTR_23";
inline const OString HUSTR_24 = "HUSTR_24";
inline const OString HUSTR_25 = "HUSTR_25";
inline const OString HUSTR_26 = "HUSTR_26";
inline const OString HUSTR_27 = "HUSTR_27";
inline const OString HUSTR_28 = "HUSTR_28";
inline const OString HUSTR_29 = "HUSTR_29";
inline const OString HUSTR_30 = "HUSTR_30";
inline const OString HUSTR_31 = "HUSTR_31";
inline const OString HUSTR_32 = "HUSTR_32";
inline const OString PHUSTR_1 = "PHUSTR_1";
inline const OString PHUSTR_2 = "PHUSTR_2";
inline const OString PHUSTR_3 = "PHUSTR_3";
inline const OString PHUSTR_4 = "PHUSTR_4";
inline const OString PHUSTR_5 = "PHUSTR_5";
inline const OString PHUSTR_6 = "PHUSTR_6";
inline const OString PHUSTR_7 = "PHUSTR_7";
inline const OString PHUSTR_8 = "PHUSTR_8";
inline const OString PHUSTR_9 = "PHUSTR_9";
inline const OString PHUSTR_10 = "PHUSTR_10";
inline const OString PHUSTR_11 = "PHUSTR_11";
inline const OString PHUSTR_12 = "PHUSTR_12";
inline const OString PHUSTR_13 = "PHUSTR_13";
inline const OString PHUSTR_14 = "PHUSTR_14";
inline const OString PHUSTR_15 = "PHUSTR_15";
inline const OString PHUSTR_16 = "PHUSTR_16";
inline const OString PHUSTR_17 = "PHUSTR_17";
inline const OString PHUSTR_18 = "PHUSTR_18";
inline const OString PHUSTR_19 = "PHUSTR_19";
inline const OString PHUSTR_20 = "PHUSTR_20";
inline const OString PHUSTR_21 = "PHUSTR_21";
inline const OString PHUSTR_22 = "PHUSTR_22";
inline const OString PHUSTR_23 = "PHUSTR_23";
inline const OString PHUSTR_24 = "PHUSTR_24";
inline const OString PHUSTR_25 = "PHUSTR_25";
inline const OString PHUSTR_26 = "PHUSTR_26";
inline const OString PHUSTR_27 = "PHUSTR_27";
inline const OString PHUSTR_28 = "PHUSTR_28";
inline const OString PHUSTR_29 = "PHUSTR_29";
inline const OString PHUSTR_30 = "PHUSTR_30";
inline const OString PHUSTR_31 = "PHUSTR_31";
inline const OString PHUSTR_32 = "PHUSTR_32";
inline const OString THUSTR_1 = "THUSTR_1";
inline const OString THUSTR_2 = "THUSTR_2";
inline const OString THUSTR_3 = "THUSTR_3";
inline const OString THUSTR_4 = "THUSTR_4";
inline const OString THUSTR_5 = "THUSTR_5";
inline const OString THUSTR_6 = "THUSTR_6";
inline const OString THUSTR_7 = "THUSTR_7";
inline const OString THUSTR_8 = "THUSTR_8";
inline const OString THUSTR_9 = "THUSTR_9";
inline const OString THUSTR_10 = "THUSTR_10";
inline const OString THUSTR_11 = "THUSTR_11";
inline const OString THUSTR_12 = "THUSTR_12";
inline const OString THUSTR_13 = "THUSTR_13";
inline const OString THUSTR_14 = "THUSTR_14";
inline const OString THUSTR_15 = "THUSTR_15";
inline const OString THUSTR_16 = "THUSTR_16";
inline const OString THUSTR_17 = "THUSTR_17";
inline const OString THUSTR_18 = "THUSTR_18";
inline const OString THUSTR_19 = "THUSTR_19";
inline const OString THUSTR_20 = "THUSTR_20";
inline const OString THUSTR_21 = "THUSTR_21";
inline const OString THUSTR_22 = "THUSTR_22";
inline const OString THUSTR_23 = "THUSTR_23";
inline const OString THUSTR_24 = "THUSTR_24";
inline const OString THUSTR_25 = "THUSTR_25";
inline const OString THUSTR_26 = "THUSTR_26";
inline const OString THUSTR_27 = "THUSTR_27";
inline const OString THUSTR_28 = "THUSTR_28";
inline const OString THUSTR_29 = "THUSTR_29";
inline const OString THUSTR_30 = "THUSTR_30";
inline const OString THUSTR_31 = "THUSTR_31";
inline const OString THUSTR_32 = "THUSTR_32";
inline const OString HUSTR_TALKTOSELF1 = "HUSTR_TALKTOSELF1";
inline const OString HUSTR_TALKTOSELF2 = "HUSTR_TALKTOSELF2";
inline const OString HUSTR_TALKTOSELF3 = "HUSTR_TALKTOSELF3";
inline const OString HUSTR_TALKTOSELF4 = "HUSTR_TALKTOSELF4";
inline const OString HUSTR_TALKTOSELF5 = "HUSTR_TALKTOSELF5";
inline const OString HUSTR_MESSAGESENT = "HUSTR_MESSAGESENT";
inline const OString AMSTR_FOLLOWON = "AMSTR_FOLLOWON";
inline const OString AMSTR_FOLLOWOFF = "AMSTR_FOLLOWOFF";
inline const OString AMSTR_GRIDON = "AMSTR_GRIDON";
inline const OString AMSTR_GRIDOFF = "AMSTR_GRIDOFF";
inline const OString AMSTR_MARKEDSPOT = "AMSTR_MARKEDSPOT";
inline const OString AMSTR_MARKSCLEARED = "AMSTR_MARKSCLEARED";
inline const OString STSTR_MUS = "STSTR_MUS";
inline const OString STSTR_NOMUS = "STSTR_NOMUS";
inline const OString STSTR_DQDON = "STSTR_DQDON";
inline const OString STSTR_DQDOFF = "STSTR_DQDOFF";
inline const OString STSTR_KFAADDED = "STSTR_KFAADDED";
inline const OString STSTR_FAADDED = "STSTR_FAADDED";
inline const OString STSTR_NCON = "STSTR_NCON";
inline const OString STSTR_NCOFF = "STSTR_NCOFF";
inline const OString STSTR_BEHOLD = "STSTR_BEHOLD";
inline const OString STSTR_BEHOLDX = "STSTR_BEHOLDX";
inline const OString STSTR_CHOPPERS = "STSTR_CHOPPERS";
inline const OString STSTR_CLEV = "STSTR_CLEV";
inline const OString E1TEXT = "E1TEXT";
inline const OString E2TEXT = "E2TEXT";
inline const OString E3TEXT = "E3TEXT";
inline const OString E4TEXT = "E4TEXT";
inline const OString C1TEXT = "C1TEXT";
inline const OString C2TEXT = "C2TEXT";
inline const OString C3TEXT = "C3TEXT";
inline const OString C4TEXT = "C4TEXT";
inline const OString C5TEXT = "C5TEXT";
inline const OString C6TEXT = "C6TEXT";
inline const OString P1TEXT = "P1TEXT";
inline const OString P2TEXT = "P2TEXT";
inline const OString P3TEXT = "P3TEXT";
inline const OString P4TEXT = "P4TEXT";
inline const OString P5TEXT = "P5TEXT";
inline const OString P6TEXT = "P6TEXT";
inline const OString T1TEXT = "T1TEXT";
inline const OString T2TEXT = "T2TEXT";
inline const OString T3TEXT = "T3TEXT";
inline const OString T4TEXT = "T4TEXT";
inline const OString T5TEXT = "T5TEXT";
inline const OString T6TEXT = "T6TEXT";
inline const OString CC_ZOMBIE = "CC_ZOMBIE";
inline const OString CC_SHOTGUN = "CC_SHOTGUN";
inline const OString CC_HEAVY = "CC_HEAVY";
inline const OString CC_IMP = "CC_IMP";
inline const OString CC_DEMON = "CC_DEMON";
inline const OString CC_LOST = "CC_LOST";
inline const OString CC_CACO = "CC_CACO";
inline const OString CC_HELL = "CC_HELL";
inline const OString CC_BARON = "CC_BARON";
inline const OString CC_ARACH = "CC_ARACH";
inline const OString CC_PAIN = "CC_PAIN";
inline const OString CC_REVEN = "CC_REVEN";
inline const OString CC_MANCU = "CC_MANCU";
inline const OString CC_ARCH = "CC_ARCH";
inline const OString CC_SPIDER = "CC_SPIDER";
inline const OString CC_CYBER = "CC_CYBER";
inline const OString CC_HERO = "CC_HERO";
inline const OString PD_BLUEC = "PD_BLUEC";
inline const OString PD_REDC = "PD_REDC";
inline const OString PD_YELLOWC = "PD_YELLOWC";
inline const OString PD_BLUES = "PD_BLUES";
inline const OString PD_REDS = "PD_REDS";
inline const OString PD_YELLOWS = "PD_YELLOWS";
inline const OString PD_ANY = "PD_ANY";
inline const OString PD_ALL3 = "PD_ALL3";
inline const OString PD_ALL6 = "PD_ALL6";
inline const OString BGFLATE1 = "BGFLATE1";
inline const OString BGFLATE2 = "BGFLATE2";
inline const OString BGFLATE3 = "BGFLATE3";
inline const OString BGFLATE4 = "BGFLATE4";
inline const OString BGFLAT06 = "BGFLAT06";
inline const OString BGFLAT11 = "BGFLAT11";
inline const OString BGFLAT20 = "BGFLAT20";
inline const OString BGFLAT30 = "BGFLAT30";
inline const OString BGFLAT15 = "BGFLAT15";
inline const OString BGFLAT31 = "BGFLAT31";
inline const OString BGCASTCALL = "BGCASTCALL";
inline const OString TXT_FRAGLIMIT = "TXT_FRAGLIMIT";
inline const OString TXT_TIMELIMIT = "TXT_TIMELIMIT";
inline const OString SPREEKILLSELF = "SPREEKILLSELF";
inline const OString SPREEOVER = "SPREEOVER";
inline const OString SPREE5 = "SPREE5";
inline const OString SPREE10 = "SPREE10";
inline const OString SPREE15 = "SPREE15";
inline const OString SPREE20 = "SPREE20";
inline const OString SPREE25 = "SPREE25";
inline const OString MULTI2 = "MULTI2";
inline const OString MULTI3 = "MULTI3";
inline const OString MULTI4 = "MULTI4";
inline const OString MULTI5 = "MULTI5";
inline const OString OB_SUICIDE = "OB_SUICIDE";
inline const OString OB_FALLING = "OB_FALLING";
inline const OString OB_CRUSH = "OB_CRUSH";
inline const OString OB_EXIT = "OB_EXIT";
inline const OString OB_WATER = "OB_WATER";
inline const OString OB_SLIME = "OB_SLIME";
inline const OString OB_LAVA = "OB_LAVA";
inline const OString OB_BARREL = "OB_BARREL";
inline const OString OB_SPLASH = "OB_SPLASH";
inline const OString OB_R_SPLASH = "OB_R_SPLASH";
inline const OString OB_ROCKET = "OB_ROCKET";
inline const OString OB_KILLEDSELF = "OB_KILLEDSELF";
inline const OString OB_STEALTHBABY = "OB_STEALTHBABY";
inline const OString OB_STEALTHVILE = "OB_STEALTHVILE";
inline const OString OB_STEALTHBARON = "OB_STEALTHBARON";
inline const OString OB_STEALTHCACO = "OB_STEALTHCACO";
inline const OString OB_STEALTHCHAINGUY = "OB_STEALTHCHAINGUY";
inline const OString OB_STEALTHDEMON = "OB_STEALTHDEMON";
inline const OString OB_STEALTHKNIGHT = "OB_STEALTHKNIGHT";
inline const OString OB_STEALTHIMP = "OB_STEALTHIMP";
inline const OString OB_STEALTHFATSO = "OB_STEALTHFATSO";
inline const OString OB_STEALTHUNDEAD = "OB_STEALTHUNDEAD";
inline const OString OB_STEALTHSHOTGUY = "OB_STEALTHSHOTGUY";
inline const OString OB_STEALTHZOMBIE = "OB_STEALTHZOMBIE";
inline const OString OB_UNDEADHIT = "OB_UNDEADHIT";
inline const OString OB_IMPHIT = "OB_IMPHIT";
inline const OString OB_CACOHIT = "OB_CACOHIT";
inline const OString OB_DEMONHIT = "OB_DEMONHIT";
inline const OString OB_SPECTREHIT = "OB_SPECTREHIT";
inline const OString OB_BARONHIT = "OB_BARONHIT";
inline const OString OB_KNIGHTHIT = "OB_KNIGHTHIT";
inline const OString OB_ZOMBIE = "OB_ZOMBIE";
inline const OString OB_SHOTGUY = "OB_SHOTGUY";
inline const OString OB_VILE = "OB_VILE";
inline const OString OB_UNDEAD = "OB_UNDEAD";
inline const OString OB_FATSO = "OB_FATSO";
inline const OString OB_CHAINGUY = "OB_CHAINGUY";
inline const OString OB_SKULL = "OB_SKULL";
inline const OString OB_IMP = "OB_IMP";
inline const OString OB_CACO = "OB_CACO";
inline const OString OB_BARON = "OB_BARON";
inline const OString OB_KNIGHT = "OB_KNIGHT";
inline const OString OB_SPIDER = "OB_SPIDER";
inline const OString OB_BABY = "OB_BABY";
inline const OString OB_CYBORG = "OB_CYBORG";
inline const OString OB_WOLFSS = "OB_WOLFSS";
inline const OString OB_CHICKEN = "OB_CHICKEN";
inline const OString OB_BEAST = "OB_BEAST";
inline const OString OB_CLINK = "OB_CLINK";
inline const OString OB_DSPARIL1 = "OB_DSPARIL1";
inline const OString OB_DSPARIL1HIT = "OB_DSPARIL1HIT";
inline const OString OB_DSPARIL2 = "OB_DSPARIL2";
inline const OString OB_DSPARIL2HIT = "OB_DSPARIL2HIT";
inline const OString OB_HERETICIMP = "OB_HERETICIMP";
inline const OString OB_HERETICIMPHIT = "OB_HERETICIMPHIT";
inline const OString OB_IRONLICH = "OB_IRONLICH";
inline const OString OB_IRONLICHHIT = "OB_IRONLICHHIT";
inline const OString OB_BONEKNIGHT = "OB_BONEKNIGHT";
inline const OString OB_BONEKNIGHTHIT = "OB_BONEKNIGHTHIT";
inline const OString OB_MUMMY = "OB_MUMMY";
inline const OString OB_MUMMYLEADER = "OB_MUMMYLEADER";
inline const OString OB_SNAKE = "OB_SNAKE";
inline const OString OB_WIZARD = "OB_WIZARD";
inline const OString OB_WIZARDHIT = "OB_WIZARDHIT";
inline const OString OB_GENMONPROJ = "OB_GENMONPROJ";
inline const OString OB_GENMONPEW = "OB_GENMONPEW";
inline const OString OB_GENMONHIT = "OB_GENMONHIT";
inline const OString OB_GENMONBOOM = "OB_GENMONBOOM";
inline const OString OB_MPFIST = "OB_MPFIST";
inline const OString OB_MPCHAINSAW = "OB_MPCHAINSAW";
inline const OString OB_MPPISTOL = "OB_MPPISTOL";
inline const OString OB_MPSHOTGUN = "OB_MPSHOTGUN";
inline const OString OB_MPSSHOTGUN = "OB_MPSSHOTGUN";
inline const OString OB_MPCHAINGUN = "OB_MPCHAINGUN";
inline const OString OB_MPROCKET = "OB_MPROCKET";
inline const OString OB_MPR_SPLASH = "OB_MPR_SPLASH";
inline const OString OB_MPPLASMARIFLE = "OB_MPPLASMARIFLE";
inline const OString OB_MPBFG_BOOM = "OB_MPBFG_BOOM";
inline const OString OB_MPBFG_SPLASH = "OB_MPBFG_SPLASH";
inline const OString OB_MPTELEFRAG = "OB_MPTELEFRAG";
inline const OString OB_RAILGUN = "OB_RAILGUN";
inline const OString OB_KILLED = "OB_KILLED";
inline const OString OB_DEFAULT = "OB_DEFAULT";
inline const OString OB_FRIENDLY1 = "OB_FRIENDLY1";
inline const OString OB_FRIENDLY2 = "OB_FRIENDLY2";
inline const OString OB_FRIENDLY3 = "OB_FRIENDLY3";
inline const OString OB_FRIENDLY4 = "OB_FRIENDLY4";
inline const OString SAVEGAMENAME = "SAVEGAMENAME";
inline const OString STARTUP1 = "STARTUP1";
inline const OString STARTUP2 = "STARTUP2";
inline const OString STARTUP3 = "STARTUP3";
inline const OString STARTUP4 = "STARTUP4";
inline const OString STARTUP5 = "STARTUP5";
inline const OString HE1TEXT = "HE1TEXT";
inline const OString HE2TEXT = "HE2TEXT";
inline const OString HE3TEXT = "HE3TEXT";
inline const OString HE4TEXT = "HE4TEXT";
inline const OString HE5TEXT = "HE5TEXT";
inline const OString HHUSTR_E1M1 = "HHUSTR_E1M1";
inline const OString HHUSTR_E1M2 = "HHUSTR_E1M2";
inline const OString HHUSTR_E1M3 = "HHUSTR_E1M3";
inline const OString HHUSTR_E1M4 = "HHUSTR_E1M4";
inline const OString HHUSTR_E1M5 = "HHUSTR_E1M5";
inline const OString HHUSTR_E1M6 = "HHUSTR_E1M6";
inline const OString HHUSTR_E1M7 = "HHUSTR_E1M7";
inline const OString HHUSTR_E1M8 = "HHUSTR_E1M8";
inline const OString HHUSTR_E1M9 = "HHUSTR_E1M9";
inline const OString HHUSTR_E2M1 = "HHUSTR_E2M1";
inline const OString HHUSTR_E2M2 = "HHUSTR_E2M2";
inline const OString HHUSTR_E2M3 = "HHUSTR_E2M3";
inline const OString HHUSTR_E2M4 = "HHUSTR_E2M4";
inline const OString HHUSTR_E2M5 = "HHUSTR_E2M5";
inline const OString HHUSTR_E2M6 = "HHUSTR_E2M6";
inline const OString HHUSTR_E2M7 = "HHUSTR_E2M7";
inline const OString HHUSTR_E2M8 = "HHUSTR_E2M8";
inline const OString HHUSTR_E2M9 = "HHUSTR_E2M9";
inline const OString HHUSTR_E3M1 = "HHUSTR_E3M1";
inline const OString HHUSTR_E3M2 = "HHUSTR_E3M2";
inline const OString HHUSTR_E3M3 = "HHUSTR_E3M3";
inline const OString HHUSTR_E3M4 = "HHUSTR_E3M4";
inline const OString HHUSTR_E3M5 = "HHUSTR_E3M5";
inline const OString HHUSTR_E3M6 = "HHUSTR_E3M6";
inline const OString HHUSTR_E3M7 = "HHUSTR_E3M7";
inline const OString HHUSTR_E3M8 = "HHUSTR_E3M8";
inline const OString HHUSTR_E3M9 = "HHUSTR_E3M9";
inline const OString HHUSTR_E4M1 = "HHUSTR_E4M1";
inline const OString HHUSTR_E4M2 = "HHUSTR_E4M2";
inline const OString HHUSTR_E4M3 = "HHUSTR_E4M3";
inline const OString HHUSTR_E4M4 = "HHUSTR_E4M4";
inline const OString HHUSTR_E4M5 = "HHUSTR_E4M5";
inline const OString HHUSTR_E4M6 = "HHUSTR_E4M6";
inline const OString HHUSTR_E4M7 = "HHUSTR_E4M7";
inline const OString HHUSTR_E4M8 = "HHUSTR_E4M8";
inline const OString HHUSTR_E4M9 = "HHUSTR_E4M9";
inline const OString HHUSTR_E5M1 = "HHUSTR_E5M1";
inline const OString HHUSTR_E5M2 = "HHUSTR_E5M2";
inline const OString HHUSTR_E5M3 = "HHUSTR_E5M3";
inline const OString HHUSTR_E5M4 = "HHUSTR_E5M4";
inline const OString HHUSTR_E5M5 = "HHUSTR_E5M5";
inline const OString HHUSTR_E5M6 = "HHUSTR_E5M6";
inline const OString HHUSTR_E5M7 = "HHUSTR_E5M7";
inline const OString HHUSTR_E5M8 = "HHUSTR_E5M8";
inline const OString HHUSTR_E5M9 = "HHUSTR_E5M9";
inline const OString TXT_GOTBLUEKEY = "TXT_GOTBLUEKEY";
inline const OString TXT_GOTYELLOWKEY = "TXT_GOTYELLOWKEY";
inline const OString TXT_GOTGREENKEY = "TXT_GOTGREENKEY";
inline const OString TXT_ARTIHEALTH = "TXT_ARTIHEALTH";
inline const OString TXT_ARTIFLY = "TXT_ARTIFLY";
inline const OString TXT_ARTIINVULNERABILITY = "TXT_ARTIINVULNERABILITY";
inline const OString TXT_ARTITOMEOFPOWER = "TXT_ARTITOMEOFPOWER";
inline const OString TXT_ARTIINVISIBILITY = "TXT_ARTIINVISIBILITY";
inline const OString TXT_ARTIEGG = "TXT_ARTIEGG";
inline const OString TXT_ARTISUPERHEALTH = "TXT_ARTISUPERHEALTH";
inline const OString TXT_ARTITORCH = "TXT_ARTITORCH";
inline const OString TXT_ARTIFIREBOMB = "TXT_ARTIFIREBOMB";
inline const OString TXT_ARTITELEPORT = "TXT_ARTITELEPORT";
inline const OString TXT_ITEMHEALTH = "TXT_ITEMHEALTH";
inline const OString TXT_ITEMBAGOFHOLDING = "TXT_ITEMBAGOFHOLDING";
inline const OString TXT_ITEMSHIELD1 = "TXT_ITEMSHIELD1";
inline const OString TXT_ITEMSHIELD2 = "TXT_ITEMSHIELD2";
inline const OString TXT_ITEMSUPERMAP = "TXT_ITEMSUPERMAP";
inline const OString TXT_AMMOGOLDWAND1 = "TXT_AMMOGOLDWAND1";
inline const OString TXT_AMMOGOLDWAND2 = "TXT_AMMOGOLDWAND2";
inline const OString TXT_AMMOMACE1 = "TXT_AMMOMACE1";
inline const OString TXT_AMMOMACE2 = "TXT_AMMOMACE2";
inline const OString TXT_AMMOCROSSBOW1 = "TXT_AMMOCROSSBOW1";
inline const OString TXT_AMMOCROSSBOW2 = "TXT_AMMOCROSSBOW2";
inline const OString TXT_AMMOBLASTER1 = "TXT_AMMOBLASTER1";
inline const OString TXT_AMMOBLASTER2 = "TXT_AMMOBLASTER2";
inline const OString TXT_AMMOSKULLROD1 = "TXT_AMMOSKULLROD1";
inline const OString TXT_AMMOSKULLROD2 = "TXT_AMMOSKULLROD2";
inline const OString TXT_AMMOPHOENIXROD1 = "TXT_AMMOPHOENIXROD1";
inline const OString TXT_AMMOPHOENIXROD2 = "TXT_AMMOPHOENIXROD2";
inline const OString TXT_WPNMACE = "TXT_WPNMACE";
inline const OString TXT_WPNCROSSBOW = "TXT_WPNCROSSBOW";
inline const OString TXT_WPNBLASTER = "TXT_WPNBLASTER";
inline const OString TXT_WPNSKULLROD = "TXT_WPNSKULLROD";
inline const OString TXT_WPNPHOENIXROD = "TXT_WPNPHOENIXROD";
inline const OString TXT_WPNGAUNTLETS = "TXT_WPNGAUNTLETS";
inline const OString TXT_NEEDBLUEKEY = "TXT_NEEDBLUEKEY";
inline const OString TXT_NEEDGREENKEY = "TXT_NEEDGREENKEY";
inline const OString TXT_NEEDYELLOWKEY = "TXT_NEEDYELLOWKEY";
inline const OString TXT_CHEATHEALTH = "TXT_CHEATHEALTH";
inline const OString TXT_CHEATKEYS = "TXT_CHEATKEYS";
inline const OString TXT_CHEATSOUNDON = "TXT_CHEATSOUNDON";
inline const OString TXT_CHEATSOUNDOFF = "TXT_CHEATSOUNDOFF";
inline const OString TXT_CHEATIDDQD = "TXT_CHEATIDDQD";
inline const OString TXT_CHEATIDKFA = "TXT_CHEATIDKFA";
inline const OString TXT_CHEATTICKERON = "TXT_CHEATTICKERON";
inline const OString TXT_CHEATTICKEROFF = "TXT_CHEATTICKEROFF";
inline const OString TXT_CHEATARTIFACTS3 = "TXT_CHEATARTIFACTS3";
inline const OString RAVENQUITMSG = "RAVENQUITMSG";
inline const OString TXT_MANA_1 = "TXT_MANA_1";
inline const OString TXT_MANA_2 = "TXT_MANA_2";
inline const OString TXT_MANA_BOTH = "TXT_MANA_BOTH";
inline const OString TXT_KEY_STEEL = "TXT_KEY_STEEL";
inline const OString TXT_KEY_CAVE = "TXT_KEY_CAVE";
inline const OString TXT_KEY_AXE = "TXT_KEY_AXE";
inline const OString TXT_KEY_FIRE = "TXT_KEY_FIRE";
inline const OString TXT_KEY_EMERALD = "TXT_KEY_EMERALD";
inline const OString TXT_KEY_DUNGEON = "TXT_KEY_DUNGEON";
inline const OString TXT_KEY_SILVER = "TXT_KEY_SILVER";
inline const OString TXT_KEY_RUSTED = "TXT_KEY_RUSTED";
inline const OString TXT_KEY_HORN = "TXT_KEY_HORN";
inline const OString TXT_KEY_SWAMP = "TXT_KEY_SWAMP";
inline const OString TXT_KEY_CASTLE = "TXT_KEY_CASTLE";
inline const OString TXT_ARTIINVULNERABILITY2 = "TXT_ARTIINVULNERABILITY2";
inline const OString TXT_ARTISUMMON = "TXT_ARTISUMMON";
inline const OString TXT_ARTIEGG2 = "TXT_ARTIEGG2";
inline const OString TXT_ARTIPOISONBAG = "TXT_ARTIPOISONBAG";
inline const OString TXT_ARTITELEPORTOTHER = "TXT_ARTITELEPORTOTHER";
inline const OString TXT_ARTISPEED = "TXT_ARTISPEED";
inline const OString TXT_ARTIBOOSTMANA = "TXT_ARTIBOOSTMANA";
inline const OString TXT_ARTIBOOSTARMOR = "TXT_ARTIBOOSTARMOR";
inline const OString TXT_ARTIBLASTRADIUS = "TXT_ARTIBLASTRADIUS";
inline const OString TXT_ARTIHEALINGRADIUS = "TXT_ARTIHEALINGRADIUS";
inline const OString TXT_ARTIPUZZSKULL = "TXT_ARTIPUZZSKULL";
inline const OString TXT_ARTIPUZZGEMBIG = "TXT_ARTIPUZZGEMBIG";
inline const OString TXT_ARTIPUZZGEMRED = "TXT_ARTIPUZZGEMRED";
inline const OString TXT_ARTIPUZZGEMGREEN1 = "TXT_ARTIPUZZGEMGREEN1";
inline const OString TXT_ARTIPUZZGEMGREEN2 = "TXT_ARTIPUZZGEMGREEN2";
inline const OString TXT_ARTIPUZZGEMBLUE1 = "TXT_ARTIPUZZGEMBLUE1";
inline const OString TXT_ARTIPUZZGEMBLUE2 = "TXT_ARTIPUZZGEMBLUE2";
inline const OString TXT_ARTIPUZZBOOK1 = "TXT_ARTIPUZZBOOK1";
inline const OString TXT_ARTIPUZZBOOK2 = "TXT_ARTIPUZZBOOK2";
inline const OString TXT_ARTIPUZZSKULL2 = "TXT_ARTIPUZZSKULL2";
inline const OString TXT_ARTIPUZZFWEAPON = "TXT_ARTIPUZZFWEAPON";
inline const OString TXT_ARTIPUZZCWEAPON = "TXT_ARTIPUZZCWEAPON";
inline const OString TXT_ARTIPUZZMWEAPON = "TXT_ARTIPUZZMWEAPON";
inline const OString TXT_ARTIPUZZGEAR = "TXT_ARTIPUZZGEAR";
inline const OString TXT_USEPUZZLEFAILED = "TXT_USEPUZZLEFAILED";
inline const OString TXT_ARMOR1 = "TXT_ARMOR1";
inline const OString TXT_ARMOR2 = "TXT_ARMOR2";
inline const OString TXT_ARMOR3 = "TXT_ARMOR3";
inline const OString TXT_ARMOR4 = "TXT_ARMOR4";
inline const OString TXT_WEAPON_F2 = "TXT_WEAPON_F2";
inline const OString TXT_WEAPON_F3 = "TXT_WEAPON_F3";
inline const OString TXT_WEAPON_F4 = "TXT_WEAPON_F4";
inline const OString TXT_WEAPON_C2 = "TXT_WEAPON_C2";
inline const OString TXT_WEAPON_C3 = "TXT_WEAPON_C3";
inline const OString TXT_WEAPON_C4 = "TXT_WEAPON_C4";
inline const OString TXT_WEAPON_M2 = "TXT_WEAPON_M2";
inline const OString TXT_WEAPON_M3 = "TXT_WEAPON_M3";
inline const OString TXT_WEAPON_M4 = "TXT_WEAPON_M4";
inline const OString TXT_QUIETUS_PIECE = "TXT_QUIETUS_PIECE";
inline const OString TXT_WRAITHVERGE_PIECE = "TXT_WRAITHVERGE_PIECE";
inline const OString TXT_BLOODSCOURGE_PIECE = "TXT_BLOODSCOURGE_PIECE";
inline const OString BBA_BONED = "BBA_BONED";
inline const OString BBA_CASTRA = "BBA_CASTRA";
inline const OString BBA_CREAMED = "BBA_CREAMED";
inline const OString BBA_DECIMAT = "BBA_DECIMAT";
inline const OString BBA_DESTRO = "BBA_DESTRO";
inline const OString BBA_DICED = "BBA_DICED";
inline const OString BBA_DISEMBO = "BBA_DISEMBO";
inline const OString BBA_FLATTE = "BBA_FLATTE";
inline const OString BBA_JUSTICE = "BBA_JUSTICE";
inline const OString BBA_MADNESS = "BBA_MADNESS";
inline const OString BBA_KILLED = "BBA_KILLED";
inline const OString BBA_MINCMEAT = "BBA_MINCMEAT";
inline const OString BBA_MASSACR = "BBA_MASSACR";
inline const OString BBA_MUTILA = "BBA_MUTILA";
inline const OString BBA_REAMED = "BBA_REAMED";
inline const OString BBA_RIPPED = "BBA_RIPPED";
inline const OString BBA_SLAUGHT = "BBA_SLAUGHT";
inline const OString BBA_SMASHED = "BBA_SMASHED";
inline const OString BBA_SODOMIZ = "BBA_SODOMIZ";
inline const OString BBA_SPLATT = "BBA_SPLATT";
inline const OString BBA_SQUASH = "BBA_SQUASH";
inline const OString BBA_THROTTL = "BBA_THROTTL";
inline const OString BBA_WASTED = "BBA_WASTED";
inline const OString BBA_BODYBAG = "BBA_BODYBAG";
inline const OString BBA_HELL = "BBA_HELL";
inline const OString BBA_TOAST = "BBA_TOAST";
inline const OString BBA_SNUFF = "BBA_SNUFF";
inline const OString BBA_HOSED = "BBA_HOSED";
inline const OString BBA_SPRAYED = "BBA_SPRAYED";
inline const OString BBA_DOGMEAT = "BBA_DOGMEAT";
inline const OString BBA_BEATEN = "BBA_BEATEN";
inline const OString BBA_EXCREMENT = "BBA_EXCREMENT";
inline const OString BBA_HAMBURGER = "BBA_HAMBURGER";
inline const OString BBA_SCROTUM = "BBA_SCROTUM";
inline const OString BBA_POPULATION = "BBA_POPULATION";
inline const OString BBA_SUICIDE = "BBA_SUICIDE";
inline const OString BBA_DARWIN = "BBA_DARWIN";
inline const OString MUSIC_E1M1 = "MUSIC_E1M1";
inline const OString MUSIC_E1M2 = "MUSIC_E1M2";
inline const OString MUSIC_E1M3 = "MUSIC_E1M3";
inline const OString MUSIC_E1M4 = "MUSIC_E1M4";
inline const OString MUSIC_E1M5 = "MUSIC_E1M5";
inline const OString MUSIC_E1M6 = "MUSIC_E1M6";
inline const OString MUSIC_E1M7 = "MUSIC_E1M7";
inline const OString MUSIC_E1M8 = "MUSIC_E1M8";
inline const OString MUSIC_E1M9 = "MUSIC_E1M9";
inline const OString MUSIC_E2M1 = "MUSIC_E2M1";
inline const OString MUSIC_E2M2 = "MUSIC_E2M2";
inline const OString MUSIC_E2M3 = "MUSIC_E2M3";
inline const OString MUSIC_E2M4 = "MUSIC_E2M4";
inline const OString MUSIC_E2M5 = "MUSIC_E2M5";
inline const OString MUSIC_E2M6 = "MUSIC_E2M6";
inline const OString MUSIC_E2M7 = "MUSIC_E2M7";
inline const OString MUSIC_E2M8 = "MUSIC_E2M8";
inline const OString MUSIC_E2M9 = "MUSIC_E2M9";
inline const OString MUSIC_E3M1 = "MUSIC_E3M1";
inline const OString MUSIC_E3M2 = "MUSIC_E3M2";
inline const OString MUSIC_E3M3 = "MUSIC_E3M3";
inline const OString MUSIC_E3M4 = "MUSIC_E3M4";
inline const OString MUSIC_E3M5 = "MUSIC_E3M5";
inline const OString MUSIC_E3M6 = "MUSIC_E3M6";
inline const OString MUSIC_E3M7 = "MUSIC_E3M7";
inline const OString MUSIC_E3M8 = "MUSIC_E3M8";
inline const OString MUSIC_E3M9 = "MUSIC_E3M9";
inline const OString MUSIC_INTER = "MUSIC_INTER";
inline const OString MUSIC_INTRO = "MUSIC_INTRO";
inline const OString MUSIC_BUNNY = "MUSIC_BUNNY";
inline const OString MUSIC_VICTOR = "MUSIC_VICTOR";
inline const OString MUSIC_INTROA = "MUSIC_INTROA";
inline const OString MUSIC_RUNNIN = "MUSIC_RUNNIN";
inline const OString MUSIC_STALKS = "MUSIC_STALKS";
inline const OString MUSIC_COUNTD = "MUSIC_COUNTD";
inline const OString MUSIC_BETWEE = "MUSIC_BETWEE";
inline const OString MUSIC_DOOM = "MUSIC_DOOM";
inline const OString MUSIC_THE_DA = "MUSIC_THE_DA";
inline const OString MUSIC_SHAWN = "MUSIC_SHAWN";
inline const OString MUSIC_DDTBLU = "MUSIC_DDTBLU";
inline const OString MUSIC_IN_CIT = "MUSIC_IN_CIT";
inline const OString MUSIC_DEAD = "MUSIC_DEAD";
inline const OString MUSIC_STLKS2 = "MUSIC_STLKS2";
inline const OString MUSIC_THEDA2 = "MUSIC_THEDA2";
inline const OString MUSIC_DOOM2 = "MUSIC_DOOM2";
inline const OString MUSIC_DDTBL2 = "MUSIC_DDTBL2";
inline const OString MUSIC_RUNNI2 = "MUSIC_RUNNI2";
inline const OString MUSIC_DEAD2 = "MUSIC_DEAD2";
inline const OString MUSIC_STLKS3 = "MUSIC_STLKS3";
inline const OString MUSIC_ROMERO = "MUSIC_ROMERO";
inline const OString MUSIC_SHAWN2 = "MUSIC_SHAWN2";
inline const OString MUSIC_MESSAG = "MUSIC_MESSAG";
inline const OString MUSIC_COUNT2 = "MUSIC_COUNT2";
inline const OString MUSIC_DDTBL3 = "MUSIC_DDTBL3";
inline const OString MUSIC_AMPIE = "MUSIC_AMPIE";
inline const OString MUSIC_THEDA3 = "MUSIC_THEDA3";
inline const OString MUSIC_ADRIAN = "MUSIC_ADRIAN";
inline const OString MUSIC_MESSG2 = "MUSIC_MESSG2";
inline const OString MUSIC_ROMER2 = "MUSIC_ROMER2";
inline const OString MUSIC_TENSE = "MUSIC_TENSE";
inline const OString MUSIC_SHAWN3 = "MUSIC_SHAWN3";
inline const OString MUSIC_OPENIN = "MUSIC_OPENIN";
inline const OString MUSIC_EVIL = "MUSIC_EVIL";
inline const OString MUSIC_ULTIMA = "MUSIC_ULTIMA";
inline const OString MUSIC_READ_M = "MUSIC_READ_M";
inline const OString MUSIC_DM2TTL = "MUSIC_DM2TTL";
inline const OString MUSIC_DM2INT = "MUSIC_DM2INT";
inline const OString TXT_BUDDHAON = "TXT_BUDDHAON";
inline const OString TXT_BUDDHAOFF = "TXT_BUDDHAOFF";
inline const OString OSPREEENDP = "OSPREEENDP";
inline const OString OSPREEENDS = "OSPREEENDS";
inline const OString OSPREEENDM = "OSPREEENDM";
inline const OString OSPREEREP = "OSPREEREP";
inline const OString OSPREE1 = "OSPREE1";
inline const OString OSPREE2 = "OSPREE2";
inline const OString OSPREE3 = "OSPREE3";
inline const OString OSPREE4 = "OSPREE4";
inline const OString OSPREE5 = "OSPREE5";
inline const OString OSPREE6 = "OSPREE6";
inline const OString OBSPREE1 = "OBSPREE1";
inline const OString OBSPREE2 = "OBSPREE2";
inline const OString OBSPREE3 = "OBSPREE3";
inline const OString OBSPREE4 = "OBSPREE4";
inline const OString OBSPREE5 = "OBSPREE5";
inline const OString OBSPREE6 = "OBSPREE6";
inline const OString OMULTI2 = "OMULTI2";
inline const OString OMULTI3 = "OMULTI3";
inline const OString OMULTI4 = "OMULTI4";
inline const OString OMULTI5 = "OMULTI5";
inline const OString OMULTI6 = "OMULTI6";
inline const OString OMULTI7 = "OMULTI7";
inline const OString OMULTI8 = "OMULTI8";
inline const OString OMULTI9 = "OMULTI9";
inline const OString OMULTI10 = "OMULTI10";
inline const OString OMULTI11 = "OMULTI11";

// Contains every original string in its proper order.
inline const OString* const stringIndexes[] = {
    &D_DEVSTR,
    &D_CDROM,
    &PRESSKEY,
    &PRESSYN,
    &QUITMSG,
    &QUITMSG1,
    &QUITMSG2,
    &QUITMSG3,
    &QUITMSG4,
    &QUITMSG5,
    &QUITMSG6,
    &QUITMSG7,
    &QUITMSG8,
    &QUITMSG9,
    &QUITMSG10,
    &QUITMSG11,
    &QUITMSG12,
    &QUITMSG13,
    &QUITMSG14,
    &LOADNET,
    &QLOADNET,
    &QSAVESPOT,
    &SAVEDEAD,
    &QSPROMPT,
    &QLPROMPT,
    &NEWGAME,
    &NIGHTMARE,
    &SWSTRING,
    &MSGOFF,
    &MSGON,
    &NETEND,
    &ENDGAME,
    &DOSY,
    &EMPTYSTRING,
    &GOTARMOR,
    &GOTMEGA,
    &GOTHTHBONUS,
    &GOTARMBONUS,
    &GOTSTIM,
    &GOTMEDINEED,
    &GOTMEDIKIT,
    &GOTSUPER,
    &GOTBLUECARD,
    &GOTYELWCARD,
    &GOTREDCARD,
    &GOTBLUESKUL,
    &GOTYELWSKUL,
    &GOTREDSKUL,
    &GOTINVUL,
    &GOTBERSERK,
    &GOTINVIS,
    &GOTSUIT,
    &GOTMAP,
    &GOTVISOR,
    &GOTMSPHERE,
    &GOTCLIP,
    &GOTCLIPBOX,
    &GOTROCKET,
    &GOTROCKBOX,
    &GOTCELL,
    &GOTCELLBOX,
    &GOTSHELLS,
    &GOTSHELLBOX,
    &GOTBACKPACK,
    &GOTBFG9000,
    &GOTCHAINGUN,
    &GOTCHAINSAW,
    &GOTLAUNCHER,
    &GOTPLASMA,
    &GOTSHOTGUN,
    &GOTSHOTGUN2,
    &PD_BLUEO,
    &PD_REDO,
    &PD_YELLOWO,
    &PD_BLUEK,
    &PD_REDK,
    &PD_YELLOWK,
    &GGSAVED,
    &HUSTR_MSGU,
    &HUSTR_E1M1,
    &HUSTR_E1M2,
    &HUSTR_E1M3,
    &HUSTR_E1M4,
    &HUSTR_E1M5,
    &HUSTR_E1M6,
    &HUSTR_E1M7,
    &HUSTR_E1M8,
    &HUSTR_E1M9,
    &HUSTR_E2M1,
    &HUSTR_E2M2,
    &HUSTR_E2M3,
    &HUSTR_E2M4,
    &HUSTR_E2M5,
    &HUSTR_E2M6,
    &HUSTR_E2M7,
    &HUSTR_E2M8,
    &HUSTR_E2M9,
    &HUSTR_E3M1,
    &HUSTR_E3M2,
    &HUSTR_E3M3,
    &HUSTR_E3M4,
    &HUSTR_E3M5,
    &HUSTR_E3M6,
    &HUSTR_E3M7,
    &HUSTR_E3M8,
    &HUSTR_E3M9,
    &HUSTR_E4M1,
    &HUSTR_E4M2,
    &HUSTR_E4M3,
    &HUSTR_E4M4,
    &HUSTR_E4M5,
    &HUSTR_E4M6,
    &HUSTR_E4M7,
    &HUSTR_E4M8,
    &HUSTR_E4M9,
    &HUSTR_1,
    &HUSTR_2,
    &HUSTR_3,
    &HUSTR_4,
    &HUSTR_5,
    &HUSTR_6,
    &HUSTR_7,
    &HUSTR_8,
    &HUSTR_9,
    &HUSTR_10,
    &HUSTR_11,
    &HUSTR_12,
    &HUSTR_13,
    &HUSTR_14,
    &HUSTR_15,
    &HUSTR_16,
    &HUSTR_17,
    &HUSTR_18,
    &HUSTR_19,
    &HUSTR_20,
    &HUSTR_21,
    &HUSTR_22,
    &HUSTR_23,
    &HUSTR_24,
    &HUSTR_25,
    &HUSTR_26,
    &HUSTR_27,
    &HUSTR_28,
    &HUSTR_29,
    &HUSTR_30,
    &HUSTR_31,
    &HUSTR_32,
    &PHUSTR_1,
    &PHUSTR_2,
    &PHUSTR_3,
    &PHUSTR_4,
    &PHUSTR_5,
    &PHUSTR_6,
    &PHUSTR_7,
    &PHUSTR_8,
    &PHUSTR_9,
    &PHUSTR_10,
    &PHUSTR_11,
    &PHUSTR_12,
    &PHUSTR_13,
    &PHUSTR_14,
    &PHUSTR_15,
    &PHUSTR_16,
    &PHUSTR_17,
    &PHUSTR_18,
    &PHUSTR_19,
    &PHUSTR_20,
    &PHUSTR_21,
    &PHUSTR_22,
    &PHUSTR_23,
    &PHUSTR_24,
    &PHUSTR_25,
    &PHUSTR_26,
    &PHUSTR_27,
    &PHUSTR_28,
    &PHUSTR_29,
    &PHUSTR_30,
    &PHUSTR_31,
    &PHUSTR_32,
    &THUSTR_1,
    &THUSTR_2,
    &THUSTR_3,
    &THUSTR_4,
    &THUSTR_5,
    &THUSTR_6,
    &THUSTR_7,
    &THUSTR_8,
    &THUSTR_9,
    &THUSTR_10,
    &THUSTR_11,
    &THUSTR_12,
    &THUSTR_13,
    &THUSTR_14,
    &THUSTR_15,
    &THUSTR_16,
    &THUSTR_17,
    &THUSTR_18,
    &THUSTR_19,
    &THUSTR_20,
    &THUSTR_21,
    &THUSTR_22,
    &THUSTR_23,
    &THUSTR_24,
    &THUSTR_25,
    &THUSTR_26,
    &THUSTR_27,
    &THUSTR_28,
    &THUSTR_29,
    &THUSTR_30,
    &THUSTR_31,
    &THUSTR_32,
    &HUSTR_TALKTOSELF1,
    &HUSTR_TALKTOSELF2,
    &HUSTR_TALKTOSELF3,
    &HUSTR_TALKTOSELF4,
    &HUSTR_TALKTOSELF5,
    &HUSTR_MESSAGESENT,
    &AMSTR_FOLLOWON,
    &AMSTR_FOLLOWOFF,
    &AMSTR_GRIDON,
    &AMSTR_GRIDOFF,
    &AMSTR_MARKEDSPOT,
    &AMSTR_MARKSCLEARED,
    &STSTR_MUS,
    &STSTR_NOMUS,
    &STSTR_DQDON,
    &STSTR_DQDOFF,
    &STSTR_KFAADDED,
    &STSTR_FAADDED,
    &STSTR_NCON,
    &STSTR_NCOFF,
    &STSTR_BEHOLD,
    &STSTR_BEHOLDX,
    &STSTR_CHOPPERS,
    &STSTR_CLEV,
    &E1TEXT,
    &E2TEXT,
    &E3TEXT,
    &E4TEXT,
    &C1TEXT,
    &C2TEXT,
    &C3TEXT,
    &C4TEXT,
    &C5TEXT,
    &C6TEXT,
    &P1TEXT,
    &P2TEXT,
    &P3TEXT,
    &P4TEXT,
    &P5TEXT,
    &P6TEXT,
    &T1TEXT,
    &T2TEXT,
    &T3TEXT,
    &T4TEXT,
    &T5TEXT,
    &T6TEXT,
    &CC_ZOMBIE,
    &CC_SHOTGUN,
    &CC_HEAVY,
    &CC_IMP,
    &CC_DEMON,
    &CC_LOST,
    &CC_CACO,
    &CC_HELL,
    &CC_BARON,
    &CC_ARACH,
    &CC_PAIN,
    &CC_REVEN,
    &CC_MANCU,
    &CC_ARCH,
    &CC_SPIDER,
    &CC_CYBER,
    &CC_HERO,
    &PD_BLUEC,
    &PD_REDC,
    &PD_YELLOWC,
    &PD_BLUES,
    &PD_REDS,
    &PD_YELLOWS,
    &PD_ANY,
    &PD_ALL3,
    &PD_ALL6,
    &BGFLATE1,
    &BGFLATE2,
    &BGFLATE3,
    &BGFLATE4,
    &BGFLAT06,
    &BGFLAT11,
    &BGFLAT20,
    &BGFLAT30,
    &BGFLAT15,
    &BGFLAT31,
    &BGCASTCALL,
    &TXT_FRAGLIMIT,
    &TXT_TIMELIMIT,
    &SPREEKILLSELF,
    &SPREEOVER,
    &SPREE5,
    &SPREE10,
    &SPREE15,
    &SPREE20,
    &SPREE25,
    &MULTI2,
    &MULTI3,
    &MULTI4,
    &MULTI5,
    &OB_SUICIDE,
    &OB_FALLING,
    &OB_CRUSH,
    &OB_EXIT,
    &OB_WATER,
    &OB_SLIME,
    &OB_LAVA,
    &OB_BARREL,
    &OB_SPLASH,
    &OB_R_SPLASH,
    &OB_ROCKET,
    &OB_KILLEDSELF,
    &OB_STEALTHBABY,
    &OB_STEALTHVILE,
    &OB_STEALTHBARON,
    &OB_STEALTHCACO,
    &OB_STEALTHCHAINGUY,
    &OB_STEALTHDEMON,
    &OB_STEALTHKNIGHT,
    &OB_STEALTHIMP,
    &OB_STEALTHFATSO,
    &OB_STEALTHUNDEAD,
    &OB_STEALTHSHOTGUY,
    &OB_STEALTHZOMBIE,
    &OB_UNDEADHIT,
    &OB_IMPHIT,
    &OB_CACOHIT,
    &OB_DEMONHIT,
    &OB_SPECTREHIT,
    &OB_BARONHIT,
    &OB_KNIGHTHIT,
    &OB_ZOMBIE,
    &OB_SHOTGUY,
    &OB_VILE,
    &OB_UNDEAD,
    &OB_FATSO,
    &OB_CHAINGUY,
    &OB_SKULL,
    &OB_IMP,
    &OB_CACO,
    &OB_BARON,
    &OB_KNIGHT,
    &OB_SPIDER,
    &OB_BABY,
    &OB_CYBORG,
    &OB_WOLFSS,
    &OB_CHICKEN,
    &OB_BEAST,
    &OB_CLINK,
    &OB_DSPARIL1,
    &OB_DSPARIL1HIT,
    &OB_DSPARIL2,
    &OB_DSPARIL2HIT,
    &OB_HERETICIMP,
    &OB_HERETICIMPHIT,
    &OB_IRONLICH,
    &OB_IRONLICHHIT,
    &OB_BONEKNIGHT,
    &OB_BONEKNIGHTHIT,
    &OB_MUMMY,
    &OB_MUMMYLEADER,
    &OB_SNAKE,
    &OB_WIZARD,
    &OB_WIZARDHIT,
    &OB_GENMONPROJ,
    &OB_GENMONPEW,
    &OB_GENMONHIT,
    &OB_GENMONBOOM,
    &OB_MPFIST,
    &OB_MPCHAINSAW,
    &OB_MPPISTOL,
    &OB_MPSHOTGUN,
    &OB_MPSSHOTGUN,
    &OB_MPCHAINGUN,
    &OB_MPROCKET,
    &OB_MPR_SPLASH,
    &OB_MPPLASMARIFLE,
    &OB_MPBFG_BOOM,
    &OB_MPBFG_SPLASH,
    &OB_MPTELEFRAG,
    &OB_RAILGUN,
    &OB_KILLED,
    &OB_DEFAULT,
    &OB_FRIENDLY1,
    &OB_FRIENDLY2,
    &OB_FRIENDLY3,
    &OB_FRIENDLY4,
    &SAVEGAMENAME,
    &STARTUP1,
    &STARTUP2,
    &STARTUP3,
    &STARTUP4,
    &STARTUP5,
    &HE1TEXT,
    &HE2TEXT,
    &HE3TEXT,
    &HE4TEXT,
    &HE5TEXT,
    &HHUSTR_E1M1,
    &HHUSTR_E1M2,
    &HHUSTR_E1M3,
    &HHUSTR_E1M4,
    &HHUSTR_E1M5,
    &HHUSTR_E1M6,
    &HHUSTR_E1M7,
    &HHUSTR_E1M8,
    &HHUSTR_E1M9,
    &HHUSTR_E2M1,
    &HHUSTR_E2M2,
    &HHUSTR_E2M3,
    &HHUSTR_E2M4,
    &HHUSTR_E2M5,
    &HHUSTR_E2M6,
    &HHUSTR_E2M7,
    &HHUSTR_E2M8,
    &HHUSTR_E2M9,
    &HHUSTR_E3M1,
    &HHUSTR_E3M2,
    &HHUSTR_E3M3,
    &HHUSTR_E3M4,
    &HHUSTR_E3M5,
    &HHUSTR_E3M6,
    &HHUSTR_E3M7,
    &HHUSTR_E3M8,
    &HHUSTR_E3M9,
    &HHUSTR_E4M1,
    &HHUSTR_E4M2,
    &HHUSTR_E4M3,
    &HHUSTR_E4M4,
    &HHUSTR_E4M5,
    &HHUSTR_E4M6,
    &HHUSTR_E4M7,
    &HHUSTR_E4M8,
    &HHUSTR_E4M9,
    &HHUSTR_E5M1,
    &HHUSTR_E5M2,
    &HHUSTR_E5M3,
    &HHUSTR_E5M4,
    &HHUSTR_E5M5,
    &HHUSTR_E5M6,
    &HHUSTR_E5M7,
    &HHUSTR_E5M8,
    &HHUSTR_E5M9,
    &TXT_GOTBLUEKEY,
    &TXT_GOTYELLOWKEY,
    &TXT_GOTGREENKEY,
    &TXT_ARTIHEALTH,
    &TXT_ARTIFLY,
    &TXT_ARTIINVULNERABILITY,
    &TXT_ARTITOMEOFPOWER,
    &TXT_ARTIINVISIBILITY,
    &TXT_ARTIEGG,
    &TXT_ARTISUPERHEALTH,
    &TXT_ARTITORCH,
    &TXT_ARTIFIREBOMB,
    &TXT_ARTITELEPORT,
    &TXT_ITEMHEALTH,
    &TXT_ITEMBAGOFHOLDING,
    &TXT_ITEMSHIELD1,
    &TXT_ITEMSHIELD2,
    &TXT_ITEMSUPERMAP,
    &TXT_AMMOGOLDWAND1,
    &TXT_AMMOGOLDWAND2,
    &TXT_AMMOMACE1,
    &TXT_AMMOMACE2,
    &TXT_AMMOCROSSBOW1,
    &TXT_AMMOCROSSBOW2,
    &TXT_AMMOBLASTER1,
    &TXT_AMMOBLASTER2,
    &TXT_AMMOSKULLROD1,
    &TXT_AMMOSKULLROD2,
    &TXT_AMMOPHOENIXROD1,
    &TXT_AMMOPHOENIXROD2,
    &TXT_WPNMACE,
    &TXT_WPNCROSSBOW,
    &TXT_WPNBLASTER,
    &TXT_WPNSKULLROD,
    &TXT_WPNPHOENIXROD,
    &TXT_WPNGAUNTLETS,
    &TXT_NEEDBLUEKEY,
    &TXT_NEEDGREENKEY,
    &TXT_NEEDYELLOWKEY,
    &TXT_CHEATHEALTH,
    &TXT_CHEATKEYS,
    &TXT_CHEATSOUNDON,
    &TXT_CHEATSOUNDOFF,
    &TXT_CHEATIDDQD,
    &TXT_CHEATIDKFA,
    &TXT_CHEATTICKERON,
    &TXT_CHEATTICKEROFF,
    &TXT_CHEATARTIFACTS3,
    &RAVENQUITMSG,
    &TXT_MANA_1,
    &TXT_MANA_2,
    &TXT_MANA_BOTH,
    &TXT_KEY_STEEL,
    &TXT_KEY_CAVE,
    &TXT_KEY_AXE,
    &TXT_KEY_FIRE,
    &TXT_KEY_EMERALD,
    &TXT_KEY_DUNGEON,
    &TXT_KEY_SILVER,
    &TXT_KEY_RUSTED,
    &TXT_KEY_HORN,
    &TXT_KEY_SWAMP,
    &TXT_KEY_CASTLE,
    &TXT_ARTIINVULNERABILITY2,
    &TXT_ARTISUMMON,
    &TXT_ARTIEGG2,
    &TXT_ARTIPOISONBAG,
    &TXT_ARTITELEPORTOTHER,
    &TXT_ARTISPEED,
    &TXT_ARTIBOOSTMANA,
    &TXT_ARTIBOOSTARMOR,
    &TXT_ARTIBLASTRADIUS,
    &TXT_ARTIHEALINGRADIUS,
    &TXT_ARTIPUZZSKULL,
    &TXT_ARTIPUZZGEMBIG,
    &TXT_ARTIPUZZGEMRED,
    &TXT_ARTIPUZZGEMGREEN1,
    &TXT_ARTIPUZZGEMGREEN2,
    &TXT_ARTIPUZZGEMBLUE1,
    &TXT_ARTIPUZZGEMBLUE2,
    &TXT_ARTIPUZZBOOK1,
    &TXT_ARTIPUZZBOOK2,
    &TXT_ARTIPUZZSKULL2,
    &TXT_ARTIPUZZFWEAPON,
    &TXT_ARTIPUZZCWEAPON,
    &TXT_ARTIPUZZMWEAPON,
    &TXT_ARTIPUZZGEAR,
    &TXT_USEPUZZLEFAILED,
    &TXT_ARMOR1,
    &TXT_ARMOR2,
    &TXT_ARMOR3,
    &TXT_ARMOR4,
    &TXT_WEAPON_F2,
    &TXT_WEAPON_F3,
    &TXT_WEAPON_F4,
    &TXT_WEAPON_C2,
    &TXT_WEAPON_C3,
    &TXT_WEAPON_C4,
    &TXT_WEAPON_M2,
    &TXT_WEAPON_M3,
    &TXT_WEAPON_M4,
    &TXT_QUIETUS_PIECE,
    &TXT_WRAITHVERGE_PIECE,
    &TXT_BLOODSCOURGE_PIECE,
    &BBA_BONED,
    &BBA_CASTRA,
    &BBA_CREAMED,
    &BBA_DECIMAT,
    &BBA_DESTRO,
    &BBA_DICED,
    &BBA_DISEMBO,
    &BBA_FLATTE,
    &BBA_JUSTICE,
    &BBA_MADNESS,
    &BBA_KILLED,
    &BBA_MINCMEAT,
    &BBA_MASSACR,
    &BBA_MUTILA,
    &BBA_REAMED,
    &BBA_RIPPED,
    &BBA_SLAUGHT,
    &BBA_SMASHED,
    &BBA_SODOMIZ,
    &BBA_SPLATT,
    &BBA_SQUASH,
    &BBA_THROTTL,
    &BBA_WASTED,
    &BBA_BODYBAG,
    &BBA_HELL,
    &BBA_TOAST,
    &BBA_SNUFF,
    &BBA_HOSED,
    &BBA_SPRAYED,
    &BBA_DOGMEAT,
    &BBA_BEATEN,
    &BBA_EXCREMENT,
    &BBA_HAMBURGER,
    &BBA_SCROTUM,
    &BBA_POPULATION,
    &BBA_SUICIDE,
    &BBA_DARWIN,
    &MUSIC_E1M1,
    &MUSIC_E1M2,
    &MUSIC_E1M3,
    &MUSIC_E1M4,
    &MUSIC_E1M5,
    &MUSIC_E1M6,
    &MUSIC_E1M7,
    &MUSIC_E1M8,
    &MUSIC_E1M9,
    &MUSIC_E2M1,
    &MUSIC_E2M2,
    &MUSIC_E2M3,
    &MUSIC_E2M4,
    &MUSIC_E2M5,
    &MUSIC_E2M6,
    &MUSIC_E2M7,
    &MUSIC_E2M8,
    &MUSIC_E2M9,
    &MUSIC_E3M1,
    &MUSIC_E3M2,
    &MUSIC_E3M3,
    &MUSIC_E3M4,
    &MUSIC_E3M5,
    &MUSIC_E3M6,
    &MUSIC_E3M7,
    &MUSIC_E3M8,
    &MUSIC_E3M9,
    &MUSIC_INTER,
    &MUSIC_INTRO,
    &MUSIC_BUNNY,
    &MUSIC_VICTOR,
    &MUSIC_INTROA,
    &MUSIC_RUNNIN,
    &MUSIC_STALKS,
    &MUSIC_COUNTD,
    &MUSIC_BETWEE,
    &MUSIC_DOOM,
    &MUSIC_THE_DA,
    &MUSIC_SHAWN,
    &MUSIC_DDTBLU,
    &MUSIC_IN_CIT,
    &MUSIC_DEAD,
    &MUSIC_STLKS2,
    &MUSIC_THEDA2,
    &MUSIC_DOOM2,
    &MUSIC_DDTBL2,
    &MUSIC_RUNNI2,
    &MUSIC_DEAD2,
    &MUSIC_STLKS3,
    &MUSIC_ROMERO,
    &MUSIC_SHAWN2,
    &MUSIC_MESSAG,
    &MUSIC_COUNT2,
    &MUSIC_DDTBL3,
    &MUSIC_AMPIE,
    &MUSIC_THEDA3,
    &MUSIC_ADRIAN,
    &MUSIC_MESSG2,
    &MUSIC_ROMER2,
    &MUSIC_TENSE,
    &MUSIC_SHAWN3,
    &MUSIC_OPENIN,
    &MUSIC_EVIL,
    &MUSIC_ULTIMA,
    &MUSIC_READ_M,
    &MUSIC_DM2TTL,
    &MUSIC_DM2INT,
    &TXT_BUDDHAON,
    &TXT_BUDDHAOFF,
    &OSPREEENDP,
    &OSPREEENDS,
    &OSPREEENDM,
    &OSPREEREP,
    &OSPREE1,
    &OSPREE2,
    &OSPREE3,
    &OSPREE4,
    &OSPREE5,
    &OSPREE6,
    &OBSPREE1,
    &OBSPREE2,
    &OBSPREE3,
    &OBSPREE4,
    &OBSPREE5,
    &OBSPREE6,
    &OMULTI2,
    &OMULTI3,
    &OMULTI4,
    &OMULTI5,
    &OMULTI6,
    &OMULTI7,
    &OMULTI8,
    &OMULTI9,
    &OMULTI10,
    &OMULTI11,
};
