syntax = "proto3";

package unittest_issues;

option csharp_namespace = "UnitTest.Issues.TestProtos";

// This file is used as part of a unit test for issue 6936
// We don't need to use it, we just have to import it in both 
// "extensions_issue6936_b.proto" and "extensions_issue6936_c.proto"

import "google/protobuf/descriptor.proto";

extend google.protobuf.MessageOptions {
    string opt = 50000;
}