#include <winver.h>
#include "rc_version.h"

VS_VERSION_INFO		VERSIONINFO
  FILEVERSION		LIBMINIUPNPC_MAJOR_VERSION,LIBMINIUPNPC_MINOR_VERSION,LIBMINIUPNPC_MICRO_VERSION,0
  PRODUCTVERSION	LIBMINIUPNPC_MAJOR_VERSION,LIBMINIUPNPC_MINOR_VERSION,LIBMINIUPNPC_MICRO_VERSION,0
  FILEFLAGSMASK		VS_FFI_FILEFLAGSMASK
#ifdef _DEBUG
  FILEFLAGS		VS_FF_DEBUG
#else
  FILEFLAGS		0
#endif
  FILEOS		VOS__WINDOWS32
  FILETYPE		VFT_DLL
  FILESUBTYPE	VFT2_UNKNOWN	// not used
BEGIN
  BLOCK "StringFileInfo"
  BEGIN
    BLOCK "04090000"  /* Lang = US English, Charset = ASCII */
    BEGIN
      VALUE "FileDescription",	"MiniUPnPc library\0"
      VALUE "FileVersion",	LIBMINIUPNPC_DOTTED_VERSION "\0"
      VALUE "InternalName",	INTERNAL_NAME
      VALUE "LegalCopyright",	"Copyright (C) Thomas Bernard\0"
      VALUE "LegalTrademarks",	"\0"
      VALUE "OriginalFilename",	INTERNAL_NAME
      VALUE "ProductName",	"MiniUPnPc\0"
      VALUE "ProductVersion",	LIBMINIUPNPC_DOTTED_VERSION "\0"
      VALUE "Comments",		"For more information visit https://miniupnp.tuxfamil.org/\0"
    END
  END
  BLOCK "VarFileInfo"
  BEGIN
      VALUE "Translation", 0x0409, 0  /* US English, ASCII */
  END
END
