// Emacs style mode select   -*- C++ -*-
//-----------------------------------------------------------------------------
//
// $Id: 5a0e4e9d0685dea547f9069afc860faf3d1c67ec $
//
// Copyright (C) 1998-2006 by Randy Heit (ZDoom).
// Copyright (C) 2006-2026 by The Odamex Team.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// DESCRIPTION:
//	Compatibility hacks for older maps that won't be updated
//
//-----------------------------------------------------------------------------

#pragma once

#include "g_level.h"

struct levelcompdata_t
{
    bool reservedLineFlag = false;
};

const levelcompdata_t& P_GetLevelCompData(fhfprint_t fingerprint);