// Protocol Buffers - Google's data interchange format
// Copyright 2016 Google Inc.  All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are
// met:
//
//     * Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above
// copyright notice, this list of conditions and the following disclaimer
// in the documentation and/or other materials provided with the
// distribution.
//     * Neither the name of Google Inc. nor the names of its
// contributors may be used to endorse or promote products derived from
// this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

syntax = "proto2";

package protobuf_deprecated;
option objc_class_prefix = "Dep";

//
// This file is like unittest_deprecated_file.proto, but uses message, enum,
// enum value, and field level deprecation.
//
// The source generated from this file needs to be inspect to confirm it has
// all of the expected annotations. It also will be compiled into the unittest
// and that compile should be clean without errors.
//

// Mix of field types marked as deprecated.
message Msg1 {
  extensions 100 to max;

  optional string string_field = 1 [deprecated=true];
  required int32 int_field = 2 [deprecated=true];
  repeated fixed32 fixed_field = 3 [deprecated=true];
  optional Msg1 msg_field = 4 [deprecated=true];
}

// Mix of extension field types marked as deprecated.
extend Msg1 {
  optional string string_ext_field = 101 [deprecated=true];
  optional int32 int_ext_field = 102 [deprecated=true];
  repeated fixed32 fixed_ext_field = 103 [deprecated=true];
  optional Msg1 msg_ext_field = 104 [deprecated=true];
}

// Mix of extension field types (scoped to a message) marked as deprecated.
message Msg1A {
  extend Msg1 {
    optional string string_ext2_field = 201 [deprecated=true];
    optional int32 int_ext2_field = 202 [deprecated=true];
    repeated fixed32 fixed_ext2_field = 203 [deprecated=true];
    optional Msg1 msg_ext2_field = 204 [deprecated=true];
  }
}

// Enum value marked as deprecated.
enum Enum1 {
  ENUM1_ONE   = 1;
  ENUM1_TWO   = 2;
  ENUM1_THREE = 3 [deprecated=true];
}

// Message marked as deprecated.
message Msg2 {
  option deprecated = true;

  optional string string_field = 1;
  required int32 int_field = 2;
  repeated fixed32 fixed_field = 3;
}

// Enum marked as deprecated.
enum Enum2 {
  option deprecated = true;

  ENUM2_ONE   = 1;
  ENUM2_TWO   = 2;
  ENUM2_THREE = 3;
}
