// Emacs style mode select   -*- C++ -*- 
//-----------------------------------------------------------------------------
//
// $Id: 3a825b17043b268f52f6bdc35e66a4a5f7c88883 $
//
// Copyright (C) 1993-1996 by id Software, Inc.
// Copyright (C) 2006-2026 by The Odamex Team.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// DESCRIPTION:
//		Default Config File.
//    
//-----------------------------------------------------------------------------


#pragma once

void M_LoadDefaults(void);

void STACK_ARGS M_SaveDefaults(std::string filename = "");

std::string M_GetConfigPath(void);
std::string M_ExpandTokens(const std::string &str);
