#region Copyright notice and license
// Protocol Buffers - Google's data interchange format
// Copyright 2015 Google Inc.  All rights reserved.
// https://developers.google.com/protocol-buffers/
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are
// met:
//
//     * Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above
// copyright notice, this list of conditions and the following disclaimer
// in the documentation and/or other materials provided with the
// distribution.
//     * Neither the name of Google Inc. nor the names of its
// contributors may be used to endorse or promote products derived from
// this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#endregion

using Google.Protobuf.TestProtos;
using Google.Protobuf.WellKnownTypes;
using NUnit.Framework;

namespace Google.Protobuf.Reflection
{
    public class TypeRegistryTest
    {
        // Most of our tests use messages. Simple test that we really can use files...
        [Test]
        public void CreateWithFileDescriptor()
        {
            var registry = TypeRegistry.FromFiles(DurationReflection.Descriptor, StructReflection.Descriptor);
            AssertDescriptorPresent(registry, Duration.Descriptor);
            AssertDescriptorPresent(registry, ListValue.Descriptor);
            AssertDescriptorAbsent(registry, Timestamp.Descriptor);
        }

        [Test]
        public void TypesFromSameFile()
        {
            // Just for kicks, let's start with a nested type
            var registry = TypeRegistry.FromMessages(TestAllTypes.Types.NestedMessage.Descriptor);
            // Top-level...
            AssertDescriptorPresent(registry, TestFieldOrderings.Descriptor);
            // ... and nested (not the same as the original NestedMessage!)
            AssertDescriptorPresent(registry, TestFieldOrderings.Types.NestedMessage.Descriptor);
        }

        [Test]
        public void DependenciesAreIncluded()
        {
            var registry = TypeRegistry.FromMessages(TestAllTypes.Descriptor);
            // Direct dependencies
            AssertDescriptorPresent(registry, ImportMessage.Descriptor);
            // Public dependencies
            AssertDescriptorPresent(registry, PublicImportMessage.Descriptor);
        }

        [Test]
        public void DuplicateFiles()
        {
            // Duplicates via dependencies and simply via repetition
            var registry = TypeRegistry.FromFiles(
                UnittestProto3Reflection.Descriptor, UnittestImportProto3Reflection.Descriptor,
                TimestampReflection.Descriptor, TimestampReflection.Descriptor);
            AssertDescriptorPresent(registry, TestAllTypes.Descriptor);
            AssertDescriptorPresent(registry, ImportMessage.Descriptor);
            AssertDescriptorPresent(registry, Timestamp.Descriptor);
        }

        private static void AssertDescriptorPresent(TypeRegistry registry, MessageDescriptor descriptor)
        {
            Assert.AreSame(descriptor, registry.Find(descriptor.FullName));
        }

        private static void AssertDescriptorAbsent(TypeRegistry registry, MessageDescriptor descriptor)
        {
            Assert.IsNull(registry.Find(descriptor.FullName));
        }
    }
}
