// Emacs style mode select   -*- C++ -*-
//-----------------------------------------------------------------------------
//
// $Id: dce1c30e8962adc060eebec20ea389df17dac03a $
//
// Copyright (C) 2000-2006 by Sergey Makovkin (CSDoom .62).
// Copyright (C) 2006-2026 by The Odamex Team.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// DESCRIPTION:
//	SV_MASTER
//
//-----------------------------------------------------------------------------

#pragma once

#include "d_player.h"
#include "p_local.h"
#include "sv_main.h"
#include "c_console.h"

void SV_InitMasters();
void SV_ListMasters ();
bool SV_AddMaster(std::string_view masterip);
bool SV_RemoveMaster (const char *masterip);
void SV_UpdateMasterServers(void);
void SV_UpdateMaster(void);
void SV_ArchiveMasters(FILE *fp);
