// Emacs style mode select   -*- C++ -*-
//-----------------------------------------------------------------------------
//
// $Id: ca2e2be8a0a9140495a1f9ab487a994c8675d23f $
//
// Copyright (C) 2006-2026 by The Odamex Team.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// DESCRIPTION:
//  Utility functions
//
// AUTHORS:
//  Russell Rice (russell at odamex dot net)
//  Michael Wood (mwoodj at knology dot net)
//
//-----------------------------------------------------------------------------

#ifndef __NET_UTILS_H__
#define __NET_UTILS_H__

#include <stdint.h>
#include <string>
#include <sstream>

/**
 * odalpapi namespace.
 *
 * All code for the odamex launcher protocol api is contained within the odalpapi
 * namespace.
 */
namespace odalpapi
{

uint64_t GetMillisNow();
int32_t OdaAddrToComponents(const std::string& HostPort, std::string &AddrOut,
                            uint16_t &PortOut);
} // namespace

#endif // __NET_UTILS_H__
