//
// Rectangle drawing routines for the Fast Light Tool Kit (FLTK).
//
// Copyright 1998-2017 by Bill Spitzak and others.
//
// This library is free software. Distribution and use rights are outlined in
// the file "COPYING" which should have been included with this file.  If this
// file is missing or damaged, see the license at:
//
//     https://www.fltk.org/COPYING.php
//
// Please see the following page on how to report bugs and issues:
//
//     https://www.fltk.org/bugs.php
//

// Note:
//      This file doesn't contribute any code to the library.
//      It is kept as a placeholder for future usage.
//      This file is not compiled (as of Aug 2017)
//
//      Fl_OpenGL_Graphics_Driver is implemented in several files
//      named Fl_OpenGL_Graphics_Driver_*.cxx.
//      This may be changed in the future.

#include <FL/gl.h>
#include "Fl_OpenGL_Graphics_Driver.H"
