syntax = "proto3";

import "proto/test_service.proto";

package foo;

option php_generic_services = true;
option php_namespace = "Bar";

service OtherGreeter {
  rpc SayHello (HelloRequest) returns (HelloReply) {}
  rpc SayHelloAgain (HelloRequest) returns (HelloReply) {}
}
