#ifndef IDC_STATIC
#define IDC_STATIC (-1)
#endif

#define IDI_ICON1                               104
#define IDD_CONFIG_BOX                          106
#define IDC_APPLYBUTTON                         1000
#define IDC_VOLUMEMODEL                         1001
#define IDC_VM_LABEL                            1002
#define IDC_BANK_ID                             1003
#define IDC_INS_BANK                            1004
#define IDC_BANK_INTERNAL                       1005
#define IDC_BANK_EXTERNAL                       1006
#define IDC_BANK_PATH                           1007
#define IDC_BROWSE_BANK                         1008
#define IDC_RESTORE_DEFAULTS                    1010
#define IDC_FLAG_TREMOLO                        1011
#define IDC_FLAG_VIBRATO                        1012
#define IDC_FLAG_SOFTPAN                        1013
#define IDC_EMULATOR                            1014
#define IDC_NUM_CHIPS                           1018
#define IDC_NUM_4OPVO                           1019
#define IDC_FLAG_SCALE                          1020
#define IDC_FLAG_FULLBRIGHT                     1021
#define IDC_ABOUT                               1022
#define IDC_CHIPEMU_LABEL                       1023
#define IDC_CHIPNUM_LABEL                       1024
#define IDC_4OPSNUM_LABEL                       1025
#define IDC_RESET_SYNTH                         1026
#define IDC_BOTTOMNOTE_LABEL                    1027
#define IDC_CHANALLOC                           1028
#define IDC_CA_LABEL                            1029
#define IDC_AUDIOOUT                            1030
#define IDC_AUDIOOUT_LABEL                      1031
#define IDC_GAIN                                1032
#define IDC_GAIN_LABEL                          1033
#define IDC_DRIVERNAME                          40000
#define IDC_DRIVERDESC                          40001
