LIBRARY
; miniupnpc library
   miniupnpc

EXPORTS
; miniupnpc
   upnpDiscover
   upnpDiscoverDevice
   upnpDiscoverDevices
   upnpDiscoverAll
   freeUPNPDevlist
   parserootdesc
   UPNP_GetValidIGD
   UPNP_GetIGDFromUrl
   GetUPNPUrls
   FreeUPNPUrls
; miniwget
   miniwget
   miniwget_getaddr
; upnpcommands
   UPNP_GetTotalBytesSent
   UPNP_GetTotalBytesReceived
   UPNP_GetTotalPacketsSent
   UPNP_GetTotalPacketsReceived
   UPNP_GetStatusInfo
   UPNP_GetConnectionTypeInfo
   UPNP_GetExternalIPAddress
   UPNP_GetLinkLayerMaxBitRates
   UPNP_AddPortMapping
   UPNP_AddAnyPortMapping
   UPNP_DeletePortMapping
   UPNP_DeletePortMappingRange
   UPNP_GetPortMappingNumberOfEntries
   UPNP_GetSpecificPortMappingEntry
   UPNP_GetGenericPortMappingEntry
   UPNP_GetListOfPortMappings
   UPNP_AddPinhole
   UPNP_CheckPinholeWorking
   UPNP_UpdatePinhole
   UPNP_GetPinholePackets
   UPNP_DeletePinhole
   UPNP_GetFirewallStatus
   UPNP_GetOutboundPinholeTimeout
; upnperrors
   strupnperror
; portlistingparse
   ParsePortListing
   FreePortListing
