// Emacs style mode select   -*- C++ -*-
//-----------------------------------------------------------------------------
//
// $Id: 8d6a529c57bb0fac24368b5b0a4504fae8b349fb $
//
// Copyright (C) 2006-2026 by The Odamex Team.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// DESCRIPTION:
//  Main application sequence
//
// AUTHORS:
//  John Corrado
//  Russell Rice (russell at odamex dot net)
//  Michael Wood (mwoodj at huntsvegas dot org)
//
//-----------------------------------------------------------------------------


#pragma once

#include "odalaunch.h"

#ifdef UNIX
#undef UNIX
#include "dlg_main.h"
#define UNIX
#else
#include "dlg_main.h"
#endif

#include <wx/app.h>

class Application : public wxApp
{
public:
	bool OnInit();
	wxInt32 OnExit();

	dlgMain* MAIN_DIALOG;
};
